package ru.dbotthepony.kstarbound.world

abstract class CoordinateMapper {
	protected fun positiveModulo(a: Int, b: Int): Int {
		val result = a % b
		return if (result < 0) result + b else result
	}

	protected fun positiveModulo(a: Double, b: Int): Double {
		val result = a % b
		return if (result < 0.0) result + b else result
	}

	protected fun positiveModulo(a: Float, b: Int): Float {
		val result = a % b
		return if (result < 0f) result + b else result
	}

	abstract fun cell(value: Int): Int
	abstract fun cell(value: Double): Double
	abstract fun cell(value: Float): Float
	abstract fun chunk(value: Int): Int

	abstract fun cellToChunk(value: Int): Int
	fun cellToChunk(value: Float): Int = cellToChunk(value.toInt())
	fun cellToChunk(value: Double): Int = cellToChunk(value.toInt())

	abstract fun cellModulus(value: Int): Int

	// inside world bounds
	abstract fun inBoundsCell(value: Int): Boolean
	abstract fun inBoundsChunk(value: Int): Boolean

	// whenever provided index is legal (if world is wrapped around then provided index is legal outside world bounds)
	open fun isValidCellIndex(value: Int): Boolean = inBoundsCell(value)
	open fun isValidChunkIndex(value: Int): Boolean = inBoundsChunk(value)

	object Infinite : CoordinateMapper() {
		override fun cell(value: Int): Int = value
		override fun cell(value: Double): Double = value
		override fun cell(value: Float): Float = value
		override fun chunk(value: Int): Int = value

		override fun cellToChunk(value: Int): Int {
			return value shr CHUNK_SIZE_BITS
		}

		override fun cellModulus(value: Int): Int {
			return value and CHUNK_SIZE_MASK
		}

		override fun inBoundsCell(value: Int) = true
		override fun inBoundsChunk(value: Int) = true
	}

	class Wrapper(private val cells: Int, private val chunkSize: Int) : CoordinateMapper() {
		private val chunks = cells / chunkSize

		override fun inBoundsCell(value: Int) = value in 0 until cells
		override fun inBoundsChunk(value: Int) = value in 0 until chunks

		override fun isValidCellIndex(value: Int) = true
		override fun isValidChunkIndex(value: Int) = true

		override fun cellToChunk(value: Int): Int {
			return chunk(value / chunkSize)
		}

		override fun cellModulus(value: Int): Int {
			return positiveModulo(value, chunkSize)
		}

		override fun cell(value: Int): Int = positiveModulo(value, cells)
		override fun cell(value: Double): Double = positiveModulo(value, cells)
		override fun cell(value: Float): Float = positiveModulo(value, cells)
		override fun chunk(value: Int): Int = positiveModulo(value, chunks)
	}

	class Clamper(private val cells: Int, private val chunkSize: Int) : CoordinateMapper() {
		private val chunks = cells / chunkSize

		override fun inBoundsCell(value: Int): Boolean {
			return value in 0 until cells
		}

		override fun cellModulus(value: Int): Int {
			return positiveModulo(value, chunkSize)
		}

		override fun cellToChunk(value: Int): Int {
			return chunk(value / chunkSize)
		}

		override fun inBoundsChunk(value: Int): Boolean {
			return value in 0 until chunks
		}

		override fun cell(value: Int): Int {
			return value.coerceIn(0, cells - 1)
		}

		override fun cell(value: Double): Double {
			return value.coerceIn(0.0, cells - 1.0)
		}

		override fun cell(value: Float): Float {
			return value.coerceIn(0f, cells - 1f)
		}

		override fun chunk(value: Int): Int {
			return value.coerceIn(0, chunks - 1)
		}
	}
}
