package ru.dbotthepony.kstarbound.world

import com.google.common.base.Predicate
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import it.unimi.dsi.fastutil.objects.ReferenceOpenHashSet
import ru.dbotthepony.kbox2d.api.ContactImpulse
import ru.dbotthepony.kbox2d.api.IContactFilter
import ru.dbotthepony.kbox2d.api.IContactListener
import ru.dbotthepony.kbox2d.api.Manifold
import ru.dbotthepony.kbox2d.dynamics.B2Fixture
import ru.dbotthepony.kbox2d.dynamics.B2World
import ru.dbotthepony.kbox2d.dynamics.contact.AbstractContact
import ru.dbotthepony.kstarbound.math.*
import ru.dbotthepony.kstarbound.util.Timer
import ru.dbotthepony.kstarbound.world.api.ICellAccess
import ru.dbotthepony.kstarbound.world.api.IChunkCell
import ru.dbotthepony.kstarbound.world.api.TileView
import ru.dbotthepony.kstarbound.world.entities.Entity
import ru.dbotthepony.kvector.api.IStruct2d
import ru.dbotthepony.kvector.api.IStruct2i
import ru.dbotthepony.kvector.arrays.Int2DArray
import ru.dbotthepony.kvector.arrays.Object2DArray
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.util2d.AABBi
import ru.dbotthepony.kvector.vector.Vector2d
import ru.dbotthepony.kvector.vector.Vector2i
import java.lang.ref.ReferenceQueue
import java.lang.ref.WeakReference

const val CHUNK_SIZE_BITS = 5
const val CHUNK_SIZE_MASK = 1 or 2 or 4 or 8 or 16
const val CHUNK_SIZE = 1 shl CHUNK_SIZE_BITS // 32
const val CHUNK_SIZE_FF = CHUNK_SIZE - 1
const val CHUNK_SIZEd = CHUNK_SIZE.toDouble()

abstract class World<This : World<This, ChunkType>, ChunkType : Chunk<This, ChunkType>>(
	val seed: Long,
	val size: Vector2i?,
	val loopX: Boolean,
	val loopY: Boolean
) {
	private fun determineChunkSize(cells: Int): Int {
		for (i in 32 downTo 1) {
			if (cells % i == 0) {
				return i
			}
		}

		throw RuntimeException("unreachable code")
	}

	val chunkWidth = size?.x?.let(::determineChunkSize) ?: CHUNK_SIZE
	val chunkHeight = size?.y?.let(::determineChunkSize) ?: CHUNK_SIZE

	abstract inner class ChunkMap : ICellAccess {
		abstract val x: CoordinateMapper
		abstract val y: CoordinateMapper

		val background = TileView.Background(this)
		val foreground = TileView.Foreground(this)

		fun inBounds(x: Int, y: Int) = this.x.inBoundsCell(x) && this.y.inBoundsCell(y)
		fun inBounds(value: IStruct2i) = this.x.inBoundsCell(value.component1()) && this.y.inBoundsCell(value.component2())

		fun cellToChunk(x: Int, y: Int) = ChunkPos(this.x.cellToChunk(x), this.y.cellToChunk(y))
		fun cellToChunk(x: Double, y: Double) = ChunkPos(this.x.cellToChunk(x.toInt()), this.y.cellToChunk(y.toInt()))
		fun cellToChunk(value: IStruct2i) = cellToChunk(value.component1(), value.component2())
		fun cellToChunk(value: IStruct2d) = cellToChunk(value.component1(), value.component2())

		final override fun randomLongFor(x: Int, y: Int) = super.randomLongFor(x, y) xor seed

		abstract operator fun get(x: Int, y: Int): ChunkType?
		operator fun get(pos: ChunkPos) = get(pos.x, pos.y)

		abstract fun promote(self: ChunkType)

		protected val queue = ReferenceQueue<ChunkType>()

		abstract fun purge()

		protected inner class Ref(chunk: ChunkType) : WeakReference<ChunkType>(chunk, queue) {
			val pos = chunk.pos
		}

		abstract fun remove(x: Int, y: Int)

		override fun getCell(x: Int, y: Int): IChunkCell? {
			if (!this.x.isValidCellIndex(x) || !this.y.isValidCellIndex(y)) return null
			val ix = this.x.cell(x)
			val iy = this.y.cell(y)
			return get(this.x.cellToChunk(ix), this.y.cellToChunk(iy))?.getCell(this.x.cellModulus(ix), this.y.cellModulus(iy))
		}

		override fun getCellDirect(x: Int, y: Int): IChunkCell? {
			if (!this.x.inBoundsCell(x) || !this.y.inBoundsCell(y)) return null
			return getCell(x, y)
		}

		protected fun create(x: Int, y: Int): ChunkType {
			purge()
			val pos = ChunkPos(x, y)
			val chunk = chunkFactory(pos)
			val orphanedInThisChunk = ArrayList<Entity>()

			for (ent in orphanedEntities) {
				val (ex, ey) = ent.position

				if (this.x.cellToChunk(ex) == x && this.y.cellToChunk(ey) == y) {
					orphanedInThisChunk.add(ent)
				}
			}

			for (ent in orphanedInThisChunk) {
				ent.chunk = chunk
			}

			return chunk
		}
	}

	// infinite chunk map is around 30% slower than rectangular one
	inner class InfiniteChunkMap : ChunkMap() {
		private val map = Long2ObjectOpenHashMap<Any>()
		override val x = CoordinateMapper.Infinite
		override val y = CoordinateMapper.Infinite

		override fun get(x: Int, y: Int): ChunkType {
			return map[ChunkPos.toLong(x, y)]?.let {
				if (it is World<*, *>.ChunkMap.Ref) {
					it.get() as ChunkType?
				} else {
					it as ChunkType?
				}
			} ?: create(x, y).also {
				map[ChunkPos.toLong(x, y)] = Ref(it)
			}
		}

		override fun purge() {
			var next = queue.poll() as World<*, *>.ChunkMap.Ref?

			while (next != null) {
				val get = map[ChunkPos.toLong(next.pos.x, next.pos.y)]

				if (get === next) {
					map.remove(ChunkPos.toLong(next.pos.x, next.pos.y))
				}

				next = queue.poll() as World<*, *>.ChunkMap.Ref?
			}
		}

		override fun promote(self: ChunkType) {
			val (x, y) = self.pos
			val ref = map[ChunkPos.toLong(x, y)]

			if (ref !is World<*, *>.ChunkMap.Ref) {
				throw IllegalStateException("Tried to promote chunk from weak to strong storage at $x, $y; but there is $ref")
			}

			if (!(ref as World<*, ChunkType>.ChunkMap.Ref).refersTo(self)) {
				throw IllegalStateException("Tried to promote chunk from weak to strong storage at $x, $y; but it doesn't refer to valid object (self: $self, referent: ${ref.get()})")
			}

			map[ChunkPos.toLong(x, y)] = self
		}

		override fun remove(x: Int, y: Int) {
			val ref = map.remove(ChunkPos.toLong(x, y))

			if (ref is World<*, *>.ChunkMap.Ref) {
				ref.clear()
			}
		}
	}

	inner class RectChunkMap : ChunkMap() {
		val width = size!!.x
		val height = size!!.y

		override val x: CoordinateMapper = if (loopX) CoordinateMapper.Wrapper(width, chunkWidth) else CoordinateMapper.Clamper(width, chunkWidth)
		override val y: CoordinateMapper = if (loopY) CoordinateMapper.Wrapper(height, chunkHeight) else CoordinateMapper.Clamper(height, chunkHeight)

		private val map = Object2DArray.nulls<Any>(width / chunkWidth, height / chunkHeight)

		override fun get(x: Int, y: Int): ChunkType? {
			if (!this.x.isValidChunkIndex(x) || !this.y.isValidChunkIndex(y)) return null
			val ix = this.x.chunk(x)
			val iy = this.x.chunk(y)

			return map[ix, iy]?.let {
				if (it is World<*, *>.ChunkMap.Ref) {
					it.get() as ChunkType?
				} else {
					it as ChunkType?
				}
			} ?: create(ix, iy).also {
				map[ix, iy] = Ref(it)
			}
		}

		override fun purge() {
			while (queue.poll() != null) {}
		}

		override fun promote(self: ChunkType) {
			val (x, y) = self.pos

			val ref = map[x, y]

			if (ref !is World<*, *>.ChunkMap.Ref) {
				throw IllegalStateException("Tried to promote chunk from weak to strong storage at $x, $y; but there is $ref")
			}

			if (!(ref as World<*, ChunkType>.ChunkMap.Ref).refersTo(self)) {
				throw IllegalStateException("Tried to promote chunk from weak to strong storage at $x, $y; but it doesn't refer to valid object (self: $self, referent: ${ref.get()})")
			}

			map[x, y] = self
		}

		override fun remove(x: Int, y: Int) {
			if (!this.x.isValidChunkIndex(x) || !this.y.isValidChunkIndex(y)) return

			val ix = this.x.chunk(x)
			val iy = this.x.chunk(y)

			val old = map[ix, iy]

			if (old is World<*, *>.ChunkMap.Ref) {
				old.clear()
			}

			map[ix, iy] = null
		}
	}

	val chunkMap: ChunkMap = if (size != null) RectChunkMap() else InfiniteChunkMap()

	/**
	 * Chunks, which have their collision mesh changed
	 */
	val dirtyPhysicsChunks = ObjectOpenHashSet<ChunkType>()

	val physics = B2World(Vector2d(0.0, -EARTH_FREEFALL_ACCELERATION))

	private var timers = ArrayList<Timer>()

	fun timer(period: Double, executionTimes: Int, func: (Timer) -> Unit): Timer {
		val timer = Timer(period, executionTimes, func)
		timers.add(timer)
		return timer
	}

	init {
		physics.contactFilter = object : IContactFilter {
			override fun shouldCollide(fixtureA: B2Fixture, fixtureB: B2Fixture): Boolean {
				val dataA = fixtureA.body!!.userData
				val dataB = fixtureB.body!!.userData

				//if (dataA is AbstractProjectileMovementController && dataB is AbstractProjectileMovementController) {
				//	return false
				//}

				return true
			}
		}

		physics.contactListener = object : IContactListener {
			override fun beginContact(contact: AbstractContact) {
				val dataA = contact.fixtureA.body!!.userData
				val dataB = contact.fixtureB.body!!.userData

				if (dataA is IContactListener) {
					dataA.beginContact(contact)
				}

				if (dataB is IContactListener) {
					dataB.beginContact(contact)
				}
			}

			override fun endContact(contact: AbstractContact) {
				val dataA = contact.fixtureA.body!!.userData
				val dataB = contact.fixtureB.body!!.userData

				if (dataA is IContactListener) {
					dataA.endContact(contact)
				}

				if (dataB is IContactListener) {
					dataB.endContact(contact)
				}
			}

			override fun preSolve(contact: AbstractContact, oldManifold: Manifold) {
				val dataA = contact.fixtureA.body!!.userData
				val dataB = contact.fixtureB.body!!.userData

				if (dataA is IContactListener) {
					dataA.preSolve(contact, oldManifold)
				}

				if (dataB is IContactListener) {
					dataB.preSolve(contact, oldManifold)
				}
			}

			override fun postSolve(contact: AbstractContact, impulse: ContactImpulse) {
				val dataA = contact.fixtureA.body!!.userData
				val dataB = contact.fixtureB.body!!.userData

				if (dataA is IContactListener) {
					dataA.postSolve(contact, impulse)
				}

				if (dataB is IContactListener) {
					dataB.postSolve(contact, impulse)
				}
			}
		}
	}

	/**
	 * Таймер этого мира, в секундах.
	 *
	 * Условия:
	 * * Это значение МОЖЕТ уменьшаться, в том числе в силу того, что клиент
	 * обсчитывает мир быстрее, чем сервер, и наоборот.
	 * * Точка отсчёта не имеет смысла (по умолчанию отсчитывается от нуля).
	 * * Все сущности обязаны использовать этот таймер для определения времени своих событий
	 * * Таймер НЕ МОЖЕТ возвращать разные значения в пределах одного тика.
	 * т.е. таймер должен быть обновлён только после или перед тиком всех потомков
	 * * НЕ ОБЯЗАН соответствовать течению реального времени, но как правило, соответствует ему
	 * (если позволяет скорость обсчёта потомков).
	 * * ОБЯЗАН быть сохранён при сохранении мира, и выставляться при его загрузке.
	 */
	var timer: Double = 0.0
		protected set

	/**
	 * Заставляет этот мир "думать" на заданное количество секунд.
	 * Вызывает все необходимые методы у потомков, обновляя игровое состояние.
	 */
	fun think(delta: Double) {
		require(delta > 0.0) { "Tried to update $this by $delta seconds" }

		try {
			for (chunk in dirtyPhysicsChunks) {
				chunk.bakeCollisions()
			}

			dirtyPhysicsChunks.clear()

			//physics.step(delta, 6, 4)

			timer += delta

			if (timers.isNotEmpty()) {
				val current = timers
				timers = ArrayList()

				val iterator = current.iterator()

				for (timer in iterator) {
					timer.think(delta)

					if (timer.destroyed) {
						iterator.remove()
					}
				}

				if (timers.isNotEmpty()) {
					current.addAll(timers)
				}

				timers = current
			}

			thinkInner(delta)
		} catch(err: Throwable) {
			throw RuntimeException("Ticking world $this", err)
		}
	}

	protected abstract fun thinkInner(delta: Double)

	/**
	 * Сущности, у которых нет чанка, но они находятся в этом мире
	 */
	val orphanedEntities = ReferenceOpenHashSet<Entity>()

	/**
	 * Сущности, находящиеся в этом мире
	 */
	val entities = ReferenceOpenHashSet<Entity>()

	/**
	 * Стандартное ускорение свободного падения в Starbound Units/секунда^2
	 *
	 * При Vector2d.ZERO = невесомость
	 */
	var gravity by physics::gravity

	protected abstract fun chunkFactory(pos: ChunkPos): ChunkType

	fun testSpace(aabb: AABB, predicate: Predicate<IChunkCell> = Predicate { it.foreground.material != null }): Boolean {
		val tiles = aabb.encasingIntAABB()

		for (x in tiles.mins.x .. tiles.maxs.x) {
			for (y in tiles.mins.y .. tiles.maxs.y) {
				if (predicate.test(chunkMap.getCell(x, y) ?: continue)) {
					return true
				}
			}
		}

		return false
	}

	// Свет

	private fun floodLightInto(
		lightmap: Int2DArray,
		thisIntensity: Int,
		lightBlockerStrength: Int,
		posX: Int,
		worldPosX: Int,
		posY: Int,
		worldPosY: Int,
	): Int {
		if (lightmap[posX, posY] >= thisIntensity) {
			return 1
		}

		val tile = chunkMap.getCell(worldPosX, worldPosY)

		var newIntensity: Int

		if (tile?.foreground?.material?.renderParameters?.lightTransparent == false) {
			newIntensity = thisIntensity - lightBlockerStrength
		} else {
			newIntensity = thisIntensity - 1
		}

		if (tile?.foreground?.material != null)
			newIntensity = 0

		lightmap[posX, posY] = newIntensity.coerceAtLeast(0)

		if (newIntensity > 1) {
			var c = 1

			c += floodLightInto(
				lightmap, newIntensity, lightBlockerStrength,
				posX + 1,
				worldPosX + 1,
				posY,
				worldPosY,
			)

			c += floodLightInto(
				lightmap, newIntensity, lightBlockerStrength,
				posX - 1,
				worldPosX - 1,
				posY,
				worldPosY,
			)

			c += floodLightInto(
				lightmap, newIntensity, lightBlockerStrength,
				posX,
				worldPosX,
				posY + 1,
				worldPosY + 1,
			)

			c += floodLightInto(
				lightmap, newIntensity, lightBlockerStrength,
				posX,
				worldPosX,
				posY - 1,
				worldPosY - 1,
			)

			return c
		}

		return 1
	}

	/**
	 * Просчитывает распространение света во все стороны на указанной позиции (в тайлах)
	 *
	 * [lightIntensity] - максимальное расстояние, которое может пройти свет из точки своего появления.
	 * Имеет жёсткое ограничение в [CHUNK_SIZE].
	 *
	 * [lightBlockerStrength] - какова стоимость "пробития" тайла насквозь, который не пропускает свет
	 */
	fun floodLight(
		lightPosition: Vector2i,
		lightIntensity: Int,
		lightBlockerStrength: Int = 4,
	): Int2DArray {
		require(lightIntensity >= 1) { "Invalid light intensity $lightIntensity" }
		require(lightBlockerStrength >= 1) { "Invalid light blocker strength $lightBlockerStrength" }
		require(lightIntensity <= CHUNK_SIZE) { "Too intensive light! $lightIntensity" }

		val lightmap = Int2DArray.allocate(lightIntensity * 2 + 1, lightIntensity * 2 + 1)

		floodLightInto(
			lightmap,
			lightIntensity,
			lightBlockerStrength,
			lightIntensity,
			lightPosition.x,
			lightIntensity,
			lightPosition.y,
		)

		return lightmap
	}
}
